'''
bwmethods -- Bullwinkle method construction kit.

Provides functions for the construction of methods at compile time.  Many
of these are meant for overriding base class methods and manipulating their
results.  This draws etensively from Perl::Moose but adds some Python
flavour to make things fit in a Pythonic world.
'''

from __version__ import *
from bwobject import BWObject

class MethodBuilder(BWObject):
    '''
    Base class for any method building application (including the *_super
    methods of this module).  This works by creating Python code on-the-fly
    to try to offset some of the performance cost of using *_super methods.

    The base class accepts keyword arguments and passes them to __init__.
    The base class will handle:

     * require_base -- True if a base class member must exist (default
        True)

    * want_args -- True if the subclass function wants the arguments
        (default True)

     * want_result -- True if the subclass function was the result from
        the base class call as the first non-self argument (default True)

     * can_override -- True if a return value from the subclass function
        will short-circuit and/or override the base class result (default
        False)

    These are all definable in the class attributes and the constructor:

    >>> class MyObject(BWObject):
    ...     @before_super(require_base=False,
    ...                   want_result=True,
    ...                   can_override=True)
    ...     def no_base_required(self, res):
    ...         return 'Hello'
    ...

    This class must be overloaded to be used:

    >>> class MyObject(BWObject):
    ...     @MethodBuilder
    ...     def no_way(self):
    ...         pass
    ...
    Traceback (most recent call last):
        ...
    NotImplementedError: No encode defined for MethodBuilder()
    '''

    require_base = True
    want_args = True
    want_result = False
    can_override = False

    def __new__(cls, fn=None, **_kw):
        if fn is None:
            return lambda f: cls(f, **_kw)
        else:
            return super(MethodBuilder, cls).__new__(cls)

    def __init__(self, fn, **_kw):
        self.fn = fn
        self.init(**_kw)

    def init(self, require_base=None, want_args=None,
                   can_override=None, want_result=None):
        if require_base is not None:
            self.require_base = require_base
        if want_args is not None:
            self.want_args = want_args
        if want_result is not None:
            self.want_result = want_result
        if can_override is not None:
            self.can_override = can_override

    def __bindclass__(self, cls, name):
        method = self.build_wrapper(cls, name)
        method.__name__ = name
        method.__doc__ = self.fn.__doc__
        return method

    def build_wrapper(self, cls, name):
        code = ['def wrapper(_self, *_args, **_kw):']
        self.encode_wrapper(code)
        v = self.get_wrapper_dict(cls, name)
        src = '\n'.join(code)
        exec src in v
        wrapper = v.pop('wrapper')
        wrapper.__src__ = src
        return wrapper

    def get_wrapper_dict(self, cls, name):
        d = dict(cls=cls, name=name, fn=self.fn)
        if self.can_override:
            d['type_none'] = type(None)
        return d

    def encode_wrapper(self, code):
        raise NotImplementedError('No encode defined for %r' % self)

class SuperMethodBuilder(MethodBuilder):
    '''
    Base class for all super-class overriding methods.
    '''

    def __bindclass__(self, cls, name):
        if self.require_base:
            if getattr(super(cls, cls), name, None) is None:
                raise TypeError('Method %r is required in superclasses of %r' %
                    (name, cls.__name__))
        return super(SuperMethodBuilder, self).__bindclass__(cls, name)

    def encode_get_base_result(self, code):
        if self.require_base:
            code.append(
                '    res = getattr(super(cls, _self), name)(*_args, **_kw)',
            )
        else:
            code.extend((
                '    super_fn = getattr(super(cls, _self), name, None)',
                '    if super_fn is None:',
                '        res = None',
                '    else:',
                '        res = super_fn(*_args, **_kw)'
            ))

    def encode_do_fn(self, code):
        if self.can_override:
            code.append('    override = ' + self.get_fn_call())
        else:
            code.append('    ' + self.get_fn_call())

    def get_fn_call(self):
        args = ('_self',) + self.get_fn_positional_args()
        if self.want_args:
            args += ('*_args',)
        args += self.get_fn_keyword_args()
        if self.want_args:
            args += ('**_kw',)
        return 'fn(%s)' % ', '.join(args)

    def get_fn_positional_args(self):
        if self.want_result:
            return ('res',)
        else:
            return ()

    def get_fn_keyword_args(self):
        return ()

    def encode_override(self, code):
        if self.can_override:
            code.extend((
                '    if override is type_none:',
                '        return None',
                '    elif override is not None:',
                '        return override',
            ))

    def encode_return(self, code):
        code.append('    return res')

class OverrideSuperMethodBuild(SuperMethodBuilder):
    '''
    >>> class Sub(BWObject):
    ...     @override_super
    ...     def fn(self):
    ...         return 'Hello'
    ...
    Traceback (most recent call last):
        ...
    TypeError: Method 'fn' is required in superclasses of 'Sub'
    >>> class Base(BWObject):
    ...     def fn(self):
    ...         return 'Hello'
    ...
    >>> class Sub(Base):
    ...     @override_super
    ...     def fn(self):
    ...         return 'World'
    ...
    >>> b = Base()
    >>> s = Sub()
    >>> b.fn()
    'Hello'
    >>> s.fn()
    'World'
    '''

    def build_wrapper(self, cls, name):
        return self.fn
override_super = OverrideSuperMethodBuild

class BeforeSuperMethodBuilder(SuperMethodBuilder):
    '''
    >>> class Sqrt(BWObject):
    ...     def sqrt(self, arg):
    ...         return arg ** .5
    ...
    >>> class OnlyPositiveSqrt(Sqrt):
    ...     @before_super
    ...     def sqrt(self, arg):
    ...         if (arg < 0):
    ...             return 0
    ...
    >>> s = Sqrt()
    >>> ops = OnlyPositiveSqrt()
    >>> s.sqrt(-1)
    Traceback (most recent call last):
        ...
    ValueError: negative number cannot be raised to a fractional power
    >>> ops.sqrt(-1)
    0
    '''

    can_override = True

    @override_super
    def encode_wrapper(self, code):
        self.encode_do_fn(code)
        self.encode_override(code)
        self.encode_get_base_result(code)
        self.encode_return(code)
before_super = BeforeSuperMethodBuilder

class AfterSuperMethodBuilder(SuperMethodBuilder):
    '''
    >>> class Base(BWObject):
    ...     y = None
    ...     def fn(self, x):
    ...         return x * 2
    ...
    >>> class Sub(Base):
    ...     @after_super
    ...     def fn(self, x):
    ...         self.y = x
    ...         return self.y
    ...
    >>> b = Base()
    >>> s = Sub()
    >>> b.fn(5)
    10
    >>> s.fn(5)
    10
    >>> b.y
    >>> s.y
    5
    '''

    @override_super
    def encode_wrapper(self, code):
        self.encode_get_base_result(code)
        self.encode_do_fn(code)
        self.encode_override(code)
        self.encode_return(code)
after_super = AfterSuperMethodBuilder

class FollowSuperMethodBuilder(AfterSuperMethodBuilder):
    '''
    >>> class Base(BWObject):
    ...     def __init__(self, x):
    ...         self.x = x
    ...
    >>> class Sub(Base):
    ...     @follow_super
    ...     def __init__(self):
    ...         self.x *= 2
    ...
    >>> b = Base(5)
    >>> s = Sub(5)
    >>> b.x
    5
    >>> s.x
    10
    '''
    want_args = False
    can_override = False
follow_super = FollowSuperMethodBuilder

class FilterSuperMethodBuilder(AfterSuperMethodBuilder):
    '''
    >>> class Doubler(BWObject):
    ...     def apply(self, arg):
    ...         return arg * 2
    ...
    >>> class Quadrupler(Doubler):
    ...     @filter_super
    ...     def apply(self, res, arg):
    ...         return res * 2
    ...
    >>> d = Doubler()
    >>> q = Quadrupler()
    >>> d.apply(2)
    4
    >>> q.apply(2)
    8
    '''
    want_result = True
    can_override = True
filter_super = FilterSuperMethodBuilder

class AroundSuperMethodBuilder(SuperMethodBuilder):
    '''
    Provides the bound superclass function as an argument to the method.
    This allows the method to do things before and after the superclass
    method is called.

    >>> class Base(BWObject):
    ...     def fn(self, x):
    ...         return x ** 2
    ...
    >>> class Sub(Base):
    ...     @around_super
    ...     def fn(self, fn, x):
    ...         x += 1
    ...         res = fn(x * 2)
    ...         return res - 1
    ...
    >>> b = Base()
    >>> s = Sub()
    >>> b.fn(5)
    25
    >>> s.fn(5)
    143

    The method cannot ask for the result (since the superclass method is
    not called prior to calling the subclass method) nor can it override
    (since its result is always accepted):

    >>> class BadClass(BWObject):
    ...     @around_super(want_result=True)
    ...     def no_good(self):
    ...         pass
    ...
    Traceback (most recent call last):
        ...
    TypeError: Cannot get base result in around methods

    >>> class BadClass(BWObject):
    ...     @around_super(can_override=True)
    ...     def no_good(self):
    ...         pass
    ...
    Traceback (most recent call last):
        ...
    TypeError: Overrides are meaningless in around methods

    However, it is possible to indicate that the base method is not
    required:

    >>> class HandleNullBase(BWObject):
    ...     @around_super(require_base=False)
    ...     def take_it_or_leave_it(self, super_fn):
    ...         if super_fn is not None:
    ...             # Do base class stuff optionally.
    ...             super_fn()
    ...         # Carry on...
    '''

    want_result = False
    can_override = False

    @override_super
    def encode_wrapper(self, code):
        self.encode_do_fn(code)
        self.encode_return(code)

    @follow_super
    def init(self):
        if self.want_result:
            raise TypeError('Cannot get base result in around methods')
        if self.can_override:
            raise TypeError('Overrides are meaningless in around methods')

    @override_super
    def encode_do_fn(self, code):
        if self.require_base:
            code.append('    super_fn = getattr(super(cls, _self), name)')
        else:
            code.append('    super_fn = getattr(super(cls, _self), name, None)')
        code.append('    res = ' + self.get_fn_call())

    @filter_super(want_args=False)
    def get_fn_positional_args(self, res):
        return res + ('super_fn',)
around_super = AroundSuperMethodBuilder

def override_result(v, type_none=type(None)):
    '''
    Methods that return overrides cannot simply return None since that is
    what non-return-value methods also return.  As a result, if a subclass
    method wants to return None, it must return type(None) instead.  To
    avoid having to do that check manually, override_result can come to the
    rescue:

    >>> class Baseclass(BWObject):
    ...     def sometimes_override(self):
    ...         return 'Hello'
    ...
    ...     def always_override(self):
    ...         return 'World'
    ...
    >>> class Overrider(Baseclass):
    ...     def __init__(self, value):
    ...         self.value = value
    ...
    ...     @before_super
    ...     def sometimes_override(self):
    ...         return self.value
    ...
    ...     @before_super
    ...     def always_override(self):
    ...         return override_result(self.value)
    ...
    >>> print Overrider('something').sometimes_override()
    something
    >>> print Overrider('something').always_override()
    something
    >>> print Overrider(None).sometimes_override()
    Hello
    >>> print Overrider(None).always_override()
    None
    '''
    return type_none if v is None else v




'''
bwcontext -- A light-weight context manager

A context is a loose collection of information that can be stashed on the
stack to maintain related state.  Each layer of the context stack maintains
its own local storage which appears to containing sub-contexts if the
associated value is not defined at a higher level.

For example, let's create a base context and put some environment-like
information in it:

>>> from bwmember import member, into
>>> class User(BWObject):
...     username = member(str)
...     roles = member(into(tuple, list), default=())
...
...     def __installctx__(self, ctx):
...         ctx.expose(self, 'roles')
...         for role in self.roles:
...             ctx.roles[role] = True
...
>>> rootctx = BWContext('rootctx')
>>> rootctx
<rootctx>
>>> rootctx.user = User(username='superman', roles=('hero', 'flying'))
>>> rootctx.user
<rootctx/user => User(roles=('hero', 'flying'), username='superman')>
>>> 'hero' in +rootctx.user.roles
True
>>> -rootctx.user.roles.hero
True
>>> bool(-rootctx.user.roles.stealthy)
False

In this example, we define a class to contain a user record, attach it to
a context, and access the user roles via the context's reference operator
(+).  There are two operators available to dereference context lookups:

 * + (plus) will raise a KeyError if the path does not exist

 * - (minus) will return None if the path does not exist

 * ~ (tilda) will return type(None) if the path does not exist

Here they are in action:

>>> print +rootctx.var
Traceback (most recent call last):
    ...
KeyError: 'var'
>>> print -rootctx.var
None
>>> print ~rootctx.var
<type 'NoneType'>

The reason for the third tilda (~) form is to allow for cases where None is
a valid return.  type(None) is also a singleton so can be tested for using
is:

>>> ~rootctx.var is type(None)
True

Attribute-based variables cannot begin with an underbar.  Those are
reserved for BWContext internal variables.

>>> rootctx._something
Traceback (most recent call last):
    ...
AttributeError: _something

=======================
=== Naming Contexts ===
=======================

Every context has a name, though in many cases it is inherited from the
parent objects.  The first argument to BWContext's constructor may be the
name which must be either None (for anonymous subcontexts) or a string.
This ensures that some lineage can be maintained for debugging:

>>> root = BWContext('root')
>>> BWContext(BWContext(BWContext(root)))
<root{3}>

However, a name is required for root contexts:

>>> BWContext()
Traceback (most recent call last):
    ...
TypeError: name must be specified for root contexts

=============================
=== Referencing Variables ===
=============================

References can also be made in an unbound manner by referencing a context
class:

>>> user = BWContext.user
>>> roles = BWContext['user.roles']

These can then be dereferences against a context by calling the references
with the binding context:

>>> user
<*BWContext/user>
>>> +(rootctx/user)
User(roles=('hero', 'flying'), username='superman')
>>> +(rootctx/user.roles)
('hero', 'flying')
>>> +(rootctx/user['roles.flying'])
True
>>> +user
Traceback (most recent call last):
    ...
KeyError: 'user'

Only contexts may be dereferenced in this way:

>>> +([]/user)
Traceback (most recent call last):
    ...
TypeError: Ref resolution requires dict, a subclass or instance of BWContext, or None

Requesting a reference of an empty path returns the same subpath:

>>> roles[''] is roles
True
>>> bound = rootctx/roles
>>> bound[''] is bound
True

========================================
=== Dereferencing Context References ===
========================================

Although tempting, there is no automatic conversion of references into
objects in the presence of operators (like add, sub, etc).  This is to
provide a uniform approach to accessing referenced items.  The only
exception to this is the power operator due to order of precedence:

>>> ctx = BWContext('temp')
>>> ctx.x = 5
>>> ctx.x ** 2
25

No special oeprators are needed for setting or deleting though.  When
referencing attributes, only one segment is attached at a time.  When
referencing by item, the path may include '.' separators:

>>> ctx = BWContext('testctx')
>>> ctx.test.it = 5
>>> del ctx.test.it
>>> ctx.test['hello.world'] = 'something'
>>> ctx.test['hello.world']
<testctx/test.hello.world => 'something'>
>>> del ctx.test['hello.world']

References (both bound and unbound) can be rebound to other contexts
(including None) using the division operator:

 * None/ref => an unbound ref

    >>> ref = ctx.test
    >>> None/ref
    <*BWContext/test>

 * {}/ref => a bound ref to a new context with that dictionary as a base.

   >>> ref = ctx.test
   >>> dict(test='hello')/ref
   <dict/test => 'hello'>

 * ctx/ref => a ref bound to that context.  The context must be an instance
    of the equivalent unbound type of the reference:

    >>> ref = ctx.test
    >>> other = BWContext('other', test='world')
    >>> other/ref
    <other/test => 'world'>
    >>> class OtherContext(BWContext):
    ...     pass
    >>> other = OtherContext('other')
    >>> oref = other/ref
    >>> ctx/oref
    Traceback (most recent call last):
        ...
    TypeError: Ref resolution requires dict, a subclass or instance of OtherContext, or None

 * ctxclass/ref => an unbound ref to the context class specified.  The
    class must be a subclass of the the equivalent unbound type of the
    reference:

    >>> ref = ctx.test
    >>> BWContext/ref
    <*BWContext/test>
    >>> class OtherContext(BWContext):
    ...     pass
    >>> other = OtherContext('other')
    >>> oref = other/ref
    >>> BWContext/oref
    Traceback (most recent call last):
        ...
    TypeError: Ref resolution requires dict, a subclass or instance of OtherContext, or None

=====================
=== Subcontexting ===
=====================

Subcontexting is a simple matter of using the base context in the
consturctor of another context.  For our example, let's change the
contextual user in effect:

>>> subctx = BWContext(rootctx)
>>> subctx.user
<rootctx{1}/user => User(roles=('hero', 'flying'), username='superman')>
>>> subctx.user = User(username='batman', roles=('hero', 'stealthy'))
>>> subctx.user
<rootctx{1}/user => User(roles=('hero', 'stealthy'), username='batman')>
>>> 'stealthy' in +subctx.user.roles
True
>>> 'flying' in +subctx.user.roles
False
>>> -subctx.user.roles.hero
True
>>> -subctx.user.roles.stealthy
True
>>> bool(-subctx.user.roles.flying)
False

Multiple subcontexting is also legal and follows Python's general rules for
subclassing:

>>> comboctx = BWContext(rootctx, subctx)
>>> comboctx.user
<rootctx{1},rootctx{2}/user => User(roles=('hero', 'stealthy'), username='batman')>
>>> -comboctx.user.roles.hero
True
>>> -comboctx.user.roles.stealthy
True
>>> bool(-comboctx.user.roles.flying)
False

>>> root = BWContext('root', x='root_x')
>>> left = BWContext('left', root, y='left_y')
>>> left2 = BWContext(left)
>>> right = BWContext('right', root, y='right_y')
>>> bottom = BWContext('bottom', left2, right, y='sub_y')
>>> bottom.x
<bottom/x => 'root_x'>
>>> left2.x
<left{1}/x => 'root_x'>
>>> bottom.a
<bottom/a => :missing:>

Contexts can also be created using the call operator() on an existing
context.  This has the same arguments as the constructor but places the
called context as the left-most base for the next context.  For example,
this is equivalent to the above "bottom" case:

>>> bottom = left2('bottom', right, y='sub_y')
>>> bottom.x
<bottom/x => 'root_x'>

======================================
=== Throwing and Catching Contexts ===
======================================

Contexts can are BWThrowables (see bwthrowable) and can be retrieved via
the class property BWContext.CURRENT:

>>> rootctx.throw()
>>> BWContext.CURRENT is rootctx
True

To create a new context based on the the current context, use the CURRENT
class property:

>>> +BWContext.CURRENT().user
User(roles=('hero', 'flying'), username='superman')

========================
=== Deep Definitions ===
========================

Context variables can also be established at construction time.  If the
name has double-underscores, they will be converted to '.' characters.

>>> ctx = BWContext.CURRENT(user__roles__hello='world')
>>> +ctx.user.roles.hello
'world'

====================
=== Partial Keys ===
====================

In some caes, a prefix key is used to reference an object (which then might
perform additional actions on a subkey):

>>> ctx = BWContext('test')
>>> ctx['test.'] = 'hello'
>>> +ctx.test
'hello'
>>> +ctx.test.it
'hello'
>>> +ctx.test.it.world
'hello'
>>> print -ctx.other
None

Partial keys is a way of placign a context inside of another:

>>> mainctx = BWContext('main')
>>> subctx = BWContext('sub')
>>> mainctx['sub.'] = subctx
>>> subctx.hello = 'world'
>>> mainctx.sub.hello
<main/sub.hello => 'world'>

==========================
=== Deleting Variables ===
==========================

Deleting context variables has the effect of masking base context
declarations:

>>> ctx = BWContext.CURRENT()
>>> del ctx.user
>>> print -ctx.user
None

=======================
=== Dynamic getters ===
=======================

The function used to get variables from a storage or reference dictionary
is controlled by the _getter() method.  This method returns a function or
method to get the values based on a key and default value.  The base class
implementation will either:

 * If partial keys are in use in this context, a function that will scan
    those prefixes if the requested key is not in the passed storage.

 * The .get method of the storage passed.

>>> from bullwinkle import filter_super
>>> class MyContext(BWContext):
...     @filter_super
...     def _getter(self, super_getter, storage):
...         def getter(key, default=None):
...             if (isinstance(key, basestring) and
...                 key.startswith('whole_world.')):
...                 return 'hello world'
...             else:
...                 return super_getter(key, default)
...         return getter
...
>>> ctx = MyContext('test')
>>> ctx.whole_world
<test/whole_world => :missing:>
>>> ctx.whole_world.something
<test/whole_world.something => 'hello world'>
>>> sub = BWContext(ctx)
>>> sub.whole_world.something
<test{1}/whole_world.something => 'hello world'>

Note that the first example is :missing: because the test is for things
that start with "whole_world.", not just "whole_world".

Any function can be returned by .getter as long as it accepts a key and a
default and can have its __self__ attribute assigned (with the underlying
dictionary).

>>> class BadContext(BWContext):
...     def _getter(self, storage):
...         return [].__getitem__
...
>>> ctx = BadContext('bad')
>>> +ctx.test
Traceback (most recent call last):
    ...
TypeError: _getter function's __self__ must be a dict
>>> sub = BWContext(ctx)
>>> +sub.test
Traceback (most recent call last):
    ...
TypeError: _getter function's __self__ must be a dict

=========================
=== _storage and _ref ===
=========================

Internally, a context flips between _storage and _ref dictionaries
depending on whether another context is referencing this one.  The current
state can be tested by examining the context\'s __dict__.  For a new
context that is unreferenced and has nothing stored, neither _storage nor 
_ref will be available.

>>> ctx = BWContext('test')
>>> '_storage' in ctx.__dict__
False
>>> '_ref' in ctx.__dict__
False

>>> ctx.abc = 'abc, rev 1'
>>> '_storage' in ctx.__dict__
True
>>> '_ref' in ctx.__dict__
False

>>> sub = BWContext(ctx)
>>> sub.abc
<test{1}/abc => 'abc, rev 1'>
>>> '_storage' in ctx.__dict__
False
>>> '_ref' in ctx.__dict__
True

>>> ctx.abc = 'abc, rev 2'
>>> '_storage' in ctx.__dict__
True
>>> '_ref' in ctx.__dict__
False

>>> ctx.abc
<test/abc => 'abc, rev 2'>
>>> sub.abc
<test{1}/abc => 'abc, rev 1'>

'''

from __version__ import *
from bwthrowable import BWThrowable
from bwcached import cached
from bwobject import BWObject
import sys, traceback

NOT_FOUND = type(None)
DELETED = KeyError

class BWContextMeta(getattr(BWThrowable, '__metaclass__', type)):
    def __getattr__(cls, name):
        return BWUnboundContextRef(cls, name.replace('__', '_'))

    def __getitem__(cls, key):
        return BWUnboundContextRef(cls, key)

    @property
    def CURRENT(cls):
        return cls.catch()

class BWContext(BWThrowable):
    __metaclass__ = BWContextMeta
    _varkeys = ()

    def __init__(_self, _name=None, *_basectx, **_kw):
        if isinstance(_name, basestring):
            _self.__dict__.update(_name = _name, _names = (_name,),
                                                 _depths = (0,))
        elif _name is not None:
            _basectx = (_name,) + _basectx
        _self.__dict__['_basectx'] = _basectx
        if not _basectx and _self.__dict__.get('_name') is None:
            raise TypeError('name must be specified for root contexts')
        if _kw:
            for key, value in _kw.iteritems():
                # Init arguments must be strings or Python raises a
                # TypeError.
                _self[key.replace('__', '.')] = value

    def __call__(_self, _name=None, *_others, **_kw):
        return type(_self)(_name, _self, *_others, **_kw)

    @cached
    def _name(self):
        return ','.join('%s{%d}' % (name, depth)
                        for name, depth in zip(self._names, self._depths))

    @cached
    def _names(self):
        result = []
        for base in self._basectx:
            result.extend(base._names)
        return result

    @cached
    def _depths(self):
        result = []
        for base in self._basectx:
            result.extend(d + 1 for d in base._depths)
        return result

    @cached
    def _getters(self):
        if '_ref' in self.__dict__:
            return (self._ref_getfn,) + self._bro
        else:
            # We will invoke storage here to make sure a _getter of some
            # kind is created.
            return (self._storage_getfn,) + self._bro

    @cached
    def _storage(self):
        ref = self.__dict__.pop('_ref', None)
        self.__dict__.pop('_getters', None)
        if ref is None:
            return {}
        else:
            self.__dict__.pop('_ref_getfn', None)
            return dict(ref)

    @cached
    def _storage_getfn(self):
        storage = self._storage
        fn = self._getter(storage)
        if not hasattr(fn, '__self__'):
            fn.__self__ = storage
        elif not isinstance(fn.__self__, dict):
            raise TypeError('_getter function\'s __self__ must be a dict')
        return fn

    @cached
    def _ref(self):
        storage = self.__dict__.pop('_storage', None)
        self.__dict__.pop('_getters', None)
        if storage is None:
            return {}
        else:
            self.__dict__.pop('_storage_getfn', None)
            return storage

    @cached
    def _ref_getfn(self):
        ref = self._ref
        fn = self._getter(ref)
        if not hasattr(fn, '__self__'):
            fn.__self__ = ref
        elif not isinstance(fn.__self__, dict):
            raise TypeError('_getter function\'s __self__ must be a dict')
        return fn

    def _getter(self, storage):
        varkeys = self._varkeys
        if len(varkeys):
            def getter(key, default=None, NOT_FOUND=NOT_FOUND,
                            self=self, varkeys=varkeys, storage=storage):
                obj = storage.get(key, NOT_FOUND)
                subkey = None
                if obj is NOT_FOUND:
                    # We've already tested for at least one varkey.
                    for prefix, value in varkeys: # pragma: no partial
                        if key.startswith(prefix):
                            obj = value
                            subkey = key[len(prefix):]
                            break
                    else:
                        obj = default
                if getattr(obj, '__ctxproperty__', False):
                    obj = obj.__ctxproperty__(self, subkey, default)
                return obj
            return getter
        else:
            return storage.get

    @cached
    def _bro(self):
        bases = self._basectx
        if not bases:
            return ()
        elif len(bases) == 1:
            return (bases[0]._ref_getfn,) + bases[0]._bro
        else:
            return tuple(reversed(self._rbro))

    @cached
    def _rbro(self):
        bases = self._basectx
        if not bases:
            return ()
        elif len(bases) == 1:
            return bases[0]._rbro + (bases[0]._ref_getfn,)
        else:
            ctx = list(bases[-1]._rbro)
            ctx.append(bases[-1]._ref_getfn)
            found = set(id(c) for c in ctx)
            for base in reversed(bases[:-1]):
                for getter in base._rbro:
                    if id(getter) not in found:
                        ctx.append(getter)
                        found.add(id(getter))
                getter = base._ref_getfn
                if id(getter) not in found:
                    ctx.append(getter)
                    found.add(id(getter))
            return tuple(ctx)

    def get(self, key, default=None, NOT_FOUND=NOT_FOUND, DELETED=DELETED):
        # This will always have at least one getter.
        for getter in self._getters:    # pragma: no partial
            obj = getter(key, NOT_FOUND)
            if obj is not NOT_FOUND:
                break
        else:
            obj = default
        if obj is DELETED:
            obj = default
        if getattr(obj, '__ctxproperty__', False):
            obj = obj(self, None, default)
        return obj

    def __getattr__(self, name, NOT_FOUND=NOT_FOUND):
        if name.startswith('_'):
            raise AttributeError(name)
        else:
            return BWBoundContextRef(self, name.replace('__', '.'))

    def __setattr__(self, name, value):
        self[name.replace('__', '.')] = value

    def __delattr__(self, name):
        del self[name.replace('__', '.')]

    def __getitem__(self, key, NOT_FOUND=NOT_FOUND):
        return BWBoundContextRef(self, key.replace('__', '.'))

    def __setitem__(self, key, value, NOT_FOUND=NOT_FOUND):
        fn = getattr(value, '__installctx__', None)
        if fn is not None:
            del self[key]
            install_ctx = type(self)(None, self)
            fn(install_ctx[key])
            uninstall = {}
            for ikey in install_ctx._ref:
                uninstall[ikey] = self.get(ikey, DELETED)
            self._storage.update(install_ctx._ref)
            self._storage[key, 'uninstall'] = uninstall
        if isinstance(key, basestring) and key.endswith('.'):
            self.__dict__['_varkeys'] = self._varkeys + ((key, value),)
            self._storage[key[:-1]] = value
        else:
            self._storage[key] = value

    def __delitem__(self, key):
        uninstall = self.get((key, 'uninstall'), None)
        if uninstall:
            self._storage.update(uninstall)
        self._storage[key] = DELETED

    def property(self, fn):
        '''
        Declares the function to be a context property (it will be called
        when referenced).

        >>> ctx = BWContext('test')
        >>> ctx.fn = ctx.property(lambda c, s, d: 'hello')
        >>> ctx.fn
        <test/fn => 'hello'>
        '''

        fn.__ctxproperty__ = True
        return fn

    def __repr__(self):
        return '<' + self._name + '>'

    def __ctxproperty__(self, basectx, subpath, default=None):
        return self.get(subpath, default)

class BWContextRef(BWObject):
    def __pos__(self, NOT_FOUND=NOT_FOUND):
        obj = self._self(NOT_FOUND)
        if obj is NOT_FOUND:
            raise KeyError(self._path)
        else:
            return obj

    def __invert__(self):
        return self._self(type(None))

    def __neg__(self):
        return self._self()

    def __pow__(self, other):
        return self._self() ** other

    def __rdiv__(self, ctx):
        if ctx is None:
            return BWUnboundContextRef(self._ctxtype, self._path)
        elif isinstance(ctx, dict):
            return BWBoundContextRef(BWContext(type(ctx).__name__, **ctx),
                                     self._path)
        elif isinstance(ctx, self._ctxtype):
            return BWBoundContextRef(ctx, self._path)
        elif isinstance(ctx, type) and issubclass(ctx, self._ctxtype):
            return BWUnboundContextRef(ctx, self._path)
        else:
            raise TypeError(('Ref resolution requires dict, a subclass or ' +
                    'instance of %s, or None') % (self._ctxtype.__name__,))

    def _self(self, default=None):
        return default

    def property(self, fn):
        fn.__ctxproperty__ = True
        return fn

    def expose(self, obj, *members):
        for member in members:
            self[member] = self.property(lambda c, s, p: getattr(obj, member))

class BWBoundContextRef(BWContextRef):
    def __init__(self, ctx, path):
        self.__dict__['_ctx'] = ctx
        self.__dict__['_path'] = path

    def __getattr__(self, name):
        return type(self)(self._ctx, self._path + '.' + name.replace('__', '_'))

    def __setattr__(self, name, value):
        self[name.replace('__', '_')] = value

    def __delattr__(self, name):
        del self[name.replace('__', '_')]

    def __getitem__(self, subpath, NOT_FOUND=NOT_FOUND):
        if subpath:
            return type(self)(self._ctx, self._path + '.' + subpath)
        else:
            return self

    def __setitem__(self, subpath, value):
        self._ctx[self._path + '.' + subpath] = value

    def __delitem__(self, subpath):
        del self._ctx[self._path + '.' + subpath]

    def _self(self, default=None):
        return self._ctx.get(self._path, default)

    def __repr__(self):
        value = ~self
        return '<%s/%s => %s>' % (self._ctx._name, self._path,
                                  ':missing:' if value is type(None)
                                                else repr(value))

    @cached
    def _ctxtype(self):
        return type(self._ctx)

class BWUnboundContextRef(BWContextRef):
    def __init__(self, ctxtype, path):
        self.__dict__['_ctxtype'] = ctxtype
        self.__dict__['_path'] = path

    def __getattr__(self, name):
        return type(self)(self._ctxtype,
                          self._path + '.' + name.replace('__', '_'))

    def __getitem__(self, subpath):
        if subpath:
            return type(self)(self._ctxtype, self._path + '.' + subpath)
        else:
            return self

    def __repr__(self):
        return '<*%s/%s>' % (self._ctxtype.__name__, self._path)




'''
bwversion -- Manage version numbers and change logs

This module simplifies the process of specifying and comparing version
descriptors and then grouping them into change log objects.  The objects
can then be printed easily.
'''

class Version(tuple):
    '''
    Manages a version number and associated changes.

    >>> Version(0.1)
    0.1
    >>> Version('0.1.2')
    0.1.2
    >>> Version('0.1.2') > Version('0.1')
    True

    Giving Version None or another Version will simply return the passed
    value:

    >>> Version(None)
    >>> v = Version(0.5)
    >>> Version(v) is v
    True

    Versions can be freely converted to strings:

    >>> v = Version(0.5)
    >>> str(v)
    '0.5'
    >>> 'v' + v
    'v0.5'
    >>> v + ' (version)'
    '0.5 (version)'
    >>> Version(0.5) + (1,)
    0.5.1
    >>> (0, 5) + Version(1)
    0.5.1

    Special care must be given to % formatting as Version objects are
    actually tuples:

    >>> 'Version %s' % v
    Traceback (most recent call last):
        ...
    TypeError: not all arguments converted during string formatting
    >>> 'Version %s' % str(v)
    'Version 0.5'
    >>> 'Version %s' % (v,)
    'Version 0.5'
    '''

    def __new__(cls, src, *changes):
        if src is None or isinstance(src, Version):
            return src
        elif isinstance(src, basestring):
            return Version(src.split('.'), *changes)
        elif isinstance(src, (int, float)):
            return Version(str(src), *changes)
        else:
            obj = super(Version, cls).__new__(cls, src)
            obj.changes = changes
            if len(changes) == 0:
                obj.label = None
            elif len(changes) == 1:
                obj.label = changes[0]
            else:
                obj.label = changes
            return obj

    def __add__(self, other):
        if isinstance(other, basestring):
            return str(self) + other
        else:
            return type(self)(super(Version, self).__add__(other))

    def __radd__(self, other):
        if isinstance(other, basestring):
            return other + str(self)
        else:
            return Version(other.__add__(self))

    def __str__(self):
        return '.'.join(map(str, self))
    __repr__ = __str__

class ChangeLog(tuple):
    '''
    >>> ChangeLog(
    ...     Version(0.1, 'Version 0.1'),
    ...     Version(0.2, 'Change A in 0.2', 'Change B in 0.2'),
    ... )
    ...
    0.2
        * Change A in 0.2
        * Change B in 0.2
    <BLANKLINE>
    0.1
        * Version 0.1
    '''

    def __new__(cls, *items):
        return super(ChangeLog, cls).__new__(cls, sorted(items))

    def __str__(self):
        return '\n\n'.join(
            '%s\n    * %s' % (
                version,
                '<No log provided>' if version.label is None
                    else version.label if len(version.changes) == 1
                    else '\n    * '.join(map(str, version.changes))
                ) for version in reversed(sorted(self)))

    __repr__ = __str__





'''
bwcached -- Lazy loading properties

Provides simple ways to produce lazy loaded values.  These are slimmer than
the assocaited lazy member() functions of bwmember and work for many cases.
In addition, these do NOT require BWObject to work so may be preffed in
some applications.

To create a property that generates the output on first access apply
@cached to the method:

>>> class Factorial(object):
...     def __init__(self, n):
...         self.n = n
...
...     @cached
...     def value(self):
...         if self.n < 2:
...             return 1
...         else:
...             return self.n * self.next_fact.value
...
...     @cached
...     def next_fact(self):
...         return Factorial(self.n - 1)
...
>>> f = Factorial(10)
>>> f.value
3628800

========================
=== Volatile results ===
========================

Sometimes it is necessary to return a value that should not be cached.  In
these cases, simply return a cached.volatile() result:

>>> class DictLookup(object):
...     'Track a dictionary, only cache if it exists.'
...     def __init__(self, src, key, default=None):
...         self.src = src
...         self.key = key
...         self.default = default
...
...     @cached
...     def value(self):
...         'Get the value, cache if it exists'
...         print "Getting value"
...         obj = self.src.get(self.key, KeyError)
...         if obj is KeyError:
...             return cached.volatile(self.default)
...         else:
...             return obj
...
>>> d = {}
>>> dl = DictLookup(d, 'var', 'NOPE')
>>> dl.value
Getting value
'NOPE'
>>> d['var'] = 'hello world'
>>> dl.value
Getting value
'hello world'
>>> dl.value
'hello world'

=========================================
=== Accessing the underlying function ===
=========================================

In some rare cases it is helpful to access the underlying method (to access
__doc__, etc).  For @cached, simply access it through the class:

>>> DictLookup.value.__doc__
'Get the value, cache if it exists'
'''

from __version__ import *
import types, sys

class Volatile(object):
    __slots__ = ['obj']

    def __init__(self, obj):
        self.obj = obj

def cached(fn, Volailte=Volatile):
    '''
    Decorates a method that (normally) should only be called once to
    compute the value of an attribute.  Once called, unless the function
    returns a cached.volatile, the result will be stored as the attribute
    until cleared via del.

    >>> class Joiner(object):
    ...     def __init__(self, src, sep=' '):
    ...         self.src = src
    ...         self.sep = sep
    ...
    ...     @cached
    ...     def value(self):
    ...         print "Computing value"
    ...         result = self.sep.join(self.src)
    ...         if isinstance(self.src, list):
    ...             # Return volatile as list might change...
    ...             return cached.volatile(result)
    ...         else:
    ...             return result
    ...
    ...     def __str__(self):
    ...         return self.value
    ...
    >>> j = Joiner(('Hello', 'world'))
    >>> print j
    Computing value
    Hello world
    >>> print j
    Hello world
    >>> j = Joiner(['Hello', 'world'])
    >>> print j
    Computing value
    Hello world
    >>> print j
    Computing value
    Hello world
    '''

    name = fn.__name__
    def wrapper(self, target, cls=None):
        if target is None:
            return fn
        else:
            obj = fn(target)
            if type(obj) is not Volatile:
                target.__dict__[name] = obj
                return obj
            else:
                return obj.obj
    cls = type(fn.__name__,
               (object,),
               dict(__doc__=fn.__doc__, __get__=wrapper))
    return cls()
cached.volatile = Volatile

def classcached(fn, Volailte=Volatile):
    '''
    Decorates a class method that (normally) should only be called once to
    compute the value of an attribute.  Once called, unless the function
    returns a cached.volatile, the result will be stored as the attribute
    until cleared via del.

    >>> class DocLoader(object):
    ...     @classcached
    ...     def docs(cls):
    ...         print 'Loading docs'
    ...         if isinstance(cls.docsrc, (list, tuple)):
    ...             return classcached.volatile(' '.join(cls.docsrc))
    ...         else:
    ...             return cls.docsrc.read()
    ...

    >>> import StringIO
    >>> class FileDocLoader(DocLoader):
    ...     docsrc = StringIO.StringIO('Hello world')
    ...
    >>> dl = FileDocLoader()
    >>> dl.docs
    Loading docs
    'Hello world'
    >>> dl.docs
    'Hello world'
    >>> dl = FileDocLoader()
    >>> dl.docs
    'Hello world'

    >>> class JoinDocLoader(DocLoader):
    ...     docsrc = ('Hello', 'world')
    ...
    >>> dl = JoinDocLoader()
    >>> dl.docs
    Loading docs
    'Hello world'
    >>> dl.docs
    Loading docs
    'Hello world'
    >>> dl = JoinDocLoader()
    >>> dl.docs
    Loading docs
    'Hello world'
    '''

    name = fn.__name__
    def wrapper(self, target, cls=None):
        target = cls or type(target)
        obj = fn(target)
        if type(obj) is not Volatile:
            setattr(target, name, obj)
            return obj
        else:
            return obj.obj
    cls = type(fn.__name__,
               (object,),
               dict(__doc__=fn.__doc__, __get__=wrapper))
    return cls()
classcached.volatile = Volatile

def cachedmethod(fn, Volailte=Volatile):
    '''
    Decorates a function that produces a method dynamically.  This is used
    for closure-based actviities that generate inline functions based on
    parameters.  Results can still be volaitle.

    >>> class BinaryOp(object):
    ...     def __init__(self, fn):
    ...         self.fn = fn
    ...
    ...     @cachedmethod
    ...     def op(self):
    ...         'Passes integer values to fn'
    ...         print "Generating function"
    ...         if self.fn is None:
    ...             return cachedmethod.volatile(self.null_fn)
    ...         else:
    ...             return lambda self, a, b: self.fn(int(a), int(b))
    ...
    ...     def null_fn(self, a, b):
    ...         return 0
    >>> adder = BinaryOp(lambda a, b: a + b)
    >>> adder.op(3, 5)
    Generating function
    8
    >>> adder.op(3, 5)
    8
    >>> BinaryOp.op.__doc__
    'Passes integer values to fn'
    >>> anonfn = BinaryOp(None)
    >>> anonfn.op(3, 5)
    Generating function
    0
    >>> anonfn.op(3, 5)
    Generating function
    0
    >>> anonfn.fn = lambda a, b: a - b
    >>> anonfn.op(3, 5)
    Generating function
    -2
    >>> anonfn.op(3, 5)
    -2
    '''

    name = fn.__name__
    def wrapper(self, target, cls=None, MethodType=types.MethodType):
        if target is None:
            return fn
        else:
            obj = fn(target)
            if type(obj) is not Volatile:
                obj = MethodType(obj, target, type(target))
                target.__dict__[name] = obj
                return obj
            else:
                return obj.obj
    cls = type(fn.__name__,
               (object,),
               dict(__doc__=fn.__doc__, __get__=wrapper))
    return cls()
cachedmethod.volatile = Volatile





from bwcached import cachedmethod, classcachedmethod
import sys

class BWCreatable(object):
    @classmethod
    def _new(cls, *_args, **_kw):
        return super(BWCreatable, cls).__new__(cls, *_args, **_kw)

    @classcachedmethod
    def __new__(cls):
        return cls._new

class BWConvertable(BWCreatable):
    '''
    Base class for all operator superclasses.  This provides no default
    operators and cannot convert any objects other than None and itself.

    >>> print BWConvertable(None)
    None
    >>> print BWConvertable(1)
    Traceback (most recent call last):
        ...
    TypeError: Cannot construct BWConvertable from number 1
    >>> print BWConvertable('hello')
    Traceback (most recent call last):
        ...
    TypeError: Cannot construct BWConvertable from string 'hello'
    >>> print BWConvertable([])
    Traceback (most recent call last):
        ...
    TypeError: Cannot construct BWConvertable from sequence []
    >>> print BWConvertable(set())
    Traceback (most recent call last):
        ...
    TypeError: Cannot construct BWConvertable from set set([])
    >>> print BWConvertable({})
    Traceback (most recent call last):
        ...
    TypeError: Cannot construct BWConvertable from dict {}
    >>> class MyType(object):
    ...     pass
    ...
    >>> print BWConvertable(MyType())
    Traceback (most recent call last):
        ...
    TypeError: Cannot construct BWConvertable from <type 'type'>

    >>> class MyConvertable(BWConvertable):
    ...     def __init__(self, x):
    ...         print type(x)
    ...         self.x = x
    ...
    ...     @classmethod
    ...     def from_string(cls, src):
    ...         return cls._new(x=int(src))
    ...
    >>> MyConvertable('5').x
    5
    '''

    def __new__(cls, src):
        if src is None or isinstance(src, cls):
            return src
        elif isinstance(src, basestring):
            return cls.from_string(src)
        elif isinstance(src, (int, float, complex)):
            return cls.from_number(src)
        elif isinstance(src, (tuple, list)):
            return cls.from_sequence(src)
        elif isinstance(src, (frozenset, set)):
            return cls.from_set(src)
        elif isinstance(src, dict):
            return cls.from_dict(src)
        else:
            raise TypeError('Cannot construct %s from %r' %
                            (cls.__name__, src))

    @classmethod
    def from_number(cls, src):
        raise TypeError('Cannot construct %s from number %r' %
                        (cls.__name__, src))

    @classmethod
    def from_string(cls, src):
        raise TypeError('Cannot construct %s from string %r' %
                        (cls.__name__, src))

    @classmethod
    def from_sequence(cls, src):
        raise TypeError('Cannot construct %s from sequence %r' %
                        (cls.__name__, src))

    @classmethod
    def from_set(cls, src):
        raise TypeError('Cannot construct %s from set %r' %
                        (cls.__name__, src))

    @classmethod
    def from_dict(cls, src):
        raise TypeError('Cannot construct %s from dict %r' %
                        (cls.__name__, src))

class BWAdder(BWCreatable):
    def __add__(self, other):
        if not isinstance(other, self):
            other = type(self)(other)
        return type(self)(self.add(self, other))

    def __radd__(self, other):
        if not isinstance(other, self):
            other = type(self)(other)
        return type(self)(self.add(other, self))

    @cachedmethod
    def add(cls, left, right):
        return super(BWAdder, cls).__add__




'''
__changelog__ -- contains the project changelog in object form.
'''

from bwversion import Version, ChangeLog

CHANGELOG = ChangeLog(
    Version('0.3.5',
        'Fixed missing into() problem',
        ),
    Version('0.3.4',
        'Added better reference mechanics to context',
        'Added rebinding of contexts to context referecnes',
        'Added ability to set partial paths on contexts',
        'Removed classcachedmethod as no good use case emerged for it',
        ),
    Version('0.3.3',
        'Added greater support for code coverage',
        'Fixed numerous bugs via code coverage'
        ),
    Version('0.3.2',
        'Added contexts',
        ),
    Version('0.3.1',
        'Added the member attribute extensions',
        ),
    Version('0.3',
        'Added type conversion support for members',
        'Added default and built values for members',
        'Added required members',
        'Added TC global',
        ),
    Version('0.2.6',
        'Added throw() and catch() as simple functions',
        ),
    Version('0.2.5',
        'Added throwables',
        'Added changelog file',
        'Modified version to be more shell friendly',
        'Made __version__ more pydoc compatible',
        ),
    Version('0.1', 'Initial pre-release at PyOhio'),
)

__doc__ += '\n' + str(CHANGELOG)

if __name__ == '__main__':
    print CHANGELOG

